import copy
import logging
import quaternion
import numpy as np

logger = logging.getLogger("VIEWER").getChild(__name__)
PARENT_FREE_SPACE = "FreeSpace"


class CreateYaml:

    def __init__(self):
        pass

    @staticmethod
    def cnv_config_2_yaml(config_data):
        """
        Layout Config データを読み込んでYAMLデータ形式に変換する
        :param config_data: Config ファイルデータ
        :return: YAML 形式に変換したデータ
        """
        cnv_yaml = None
        try:
            cnv_yaml = {"Environment": []}
            field_body = {"name": config_data["field_body"].get("name"),
                          "parent": "World",
                          "rotation": config_data["field_body"].get("rotation"),
                          "translation": config_data["field_body"].get("translation"),
                          "uri": config_data["field_body"].get("url")
                          }
            cnv_yaml["Environment"].append(field_body)
            cnv_yaml["Environment"].append({"name": "Obstacle", "parent": "World"})
            cnv_yaml["Environment"].append({"name": "FreeSpace", "parent": "Obstacle"})

            static_body_a = config_data.get("static_bodys")
            if static_body_a is not None:
                for obstacle_body in static_body_a:
                    obstacle_body = {"name": obstacle_body.get("name"),
                                     "parent": "FreeSpace",
                                     "rotation": obstacle_body.get("rotation"),
                                     "translation": obstacle_body.get("translation"),
                                     "uri": obstacle_body.get("url")
                                     }
                    cnv_yaml["Environment"].append(obstacle_body)
        except Exception as e:
            logger.error(e)
        return cnv_yaml

    @staticmethod
    def find_body(name, body_a):
        """
        名称が一致するbody dataを返す
        :param name: 対象の名前
        :param body_a: bodyリスト
        :return: body data
        """
        body_data = None
        for body in body_a:
            if name in body.get("name"):
                body_data = body
                return body_data
        return body_data

    @classmethod
    def set_element(cls, yaml_data, object_a, body_a):
        """
        物体情報をYAML形式で配置する
        :param yaml_data: YAML data
        :param object_a: 物体情報 
        :return: 物体を配置したYAML形式データ
        """
        for index in range(len(object_a)):

            set_name = "{}-{}".format(object_a[index]["name"], index+1)
            set_parent = PARENT_FREE_SPACE
            body = cls.find_body(object_a[index]["name"], body_a)
            if body is None:
                continue
            set_uri = body.get("url")
            if set_uri is None:
                continue
            set_x = object_a[index]["position"]["x"]
            set_y = object_a[index]["position"]["y"]

            coord_z = body.get("height")
            set_translation = [set_x, set_y, coord_z]
            try:
                quat = np.quaternion(-object_a[index]["quaternion"]["w"], object_a[index]["quaternion"]["x"], object_a[index]["quaternion"]["y"], object_a[index]["quaternion"]["z"])
                euler = quaternion.as_euler_angles(quat)
                deg = float(np.rad2deg(float(euler[2])))*2
                if deg > 0:
                    deg = deg % 360
                else:
                    deg = (deg % 360) * -1
            except Exception as e:
                deg = 0.0
            set_rotation = [0.0, 0.0, deg]
            print("{}: {}".format(object_a[index]["name"],  set_rotation))
            set_data = {"name": set_name, "parent": set_parent, "uri": set_uri, "translation": list(set_translation), "rotation": set_rotation}
            yaml_data["Environment"].append(set_data)
        else:
            set_data = {"name": "end", "parent": PARENT_FREE_SPACE, "uri": "${SHARE}/LICTiA/model/dummy.body", "translation": [0,0,-1], "rotation": [0,0,0]}
            yaml_data["Environment"].append(set_data)
        return yaml_data

    @staticmethod
    def update(object_a, yaml_data, obstacles_tbl):
        """
        YAMLデータに物体情報を配置する
        :param object_a: 物体情報リスト
        :param yaml_data: Base YAML データ
        :param obstacles_tbl: 物体をYAMLデータに置き換え情報
        :return 1: 更新結果
        :return 2: 更新データ
        """
        result = False
        update_yaml_data = copy.deepcopy(yaml_data)
        try:
            if yaml_data is None:
                return result, update_yaml_data
            update_yaml_data = CreateYaml.set_element(update_yaml_data, object_a, obstacles_tbl)
            result = True
        except Exception as e:
            logger.error(e)
        return result, update_yaml_data



if __name__ == "__main__":
    set_dummy = [{"name":"Table","object_id":1,"coord":{"x":3.8130001608398256,"y":-6.793348368367841,"z":0.72},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.624888611604888,"w":0.7807139188502504}},{"name":"Table","object_id":2,"coord":{"x":3.764661804893422,"y":-4.8075594909142705,"z":0.72},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.6510776576139041,"w":0.7590111222874086}},{"name":"Table","object_id":3,"coord":{"x":5.923418155603212,"y":-4.7711079094827955,"z":0.72},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.5979416142212324,"w":0.8015396596441794}},{"name":"Table","object_id":4,"coord":{"x":6.229717575731127,"y":-6.968666381007683,"z":0.72},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.2635740445654651,"w":0.9646391672700224}},{"name":"Chair","object_id":1,"coord":{"x":3.4693877551020407,"y":-7.142857142857142,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.0,"w":1.0}},{"name":"Chair","object_id":2,"coord":{"x":3.4693877551020407,"y":-5.1020408163265305,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.7210239105529052,"w":0.6929101820661869}},{"name":"Chair","object_id":8,"coord":{"x":3.8775510204081627,"y":-5.5102040816326525,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.0,"w":1.0}},{"name":"Chair","object_id":3,"coord":{"x":4.2857142857142865,"y":-4.6938775510204085,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.6658136181659445,"w":0.746118104501408}},{"name":"Chair","object_id":4,"coord":{"x":4.6938775510204085,"y":-7.142857142857142,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.0,"w":1.0}},{"name":"Chair","object_id":5,"coord":{"x":5.1020408163265305,"y":-5.1020408163265305,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.0,"w":1.0}},{"name":"Chair","object_id":6,"coord":{"x":5.510204081632654,"y":-7.142857142857142,"z":1.099999999999996},"quaternion":{"x":-0.0,"y":-0.0,"z":-0.0,"w":1.0}}]
    res = CreateYaml.update(set_dummy, "./cnoid_item_list1.yaml", "./cnoid_item_list2.yaml")
    # pass
    # main_yaml_data = CreateYaml.read("./cnoid_item_base_list.yaml")
    # main_yaml_data = CreateYaml.del_element(main_yaml_data)
    # CreateYaml.write(main_yaml_data, "./cnoid_item_list1.yaml")
