#!/usr/bin/env python3
import copy
import os
import time

from cnoid.Util import *
from cnoid.Base import *
from cnoid.Body import *
from cnoid.BodyPlugin import *
import math
import datetime
import logging


logger = logging.getLogger("VIEWER").getChild(__name__)

STR_OBSTACLE = "Obstacle"
STR_FREE_SPACE = "FreeSpace"
STR_ENVIRONMENT = "Environment"
STR_WORLD_BODY_NAME = "World"
STR_END_BODY_NAME = "end"


class RearrangementObstacles:

    def __init__(self, base_yaml):
        self.item_list = base_yaml
        self.env_item = []
        self.world_item = WorldItem()

    @staticmethod
    def scene_view_properties():
        """
        Viewer プロパティ設定
        """
        # Set view properties.
        sceneWidget = SceneView.instance.sceneWidget
        sceneWidget.setHeadLightEnabled(True)
        sceneWidget.setHeadLightIntensity(0.1)
        sceneWidget.setFloorGridEnabled(False)
        sceneWidget.setWorldLightIntensity(0.25)
        sceneWidget.setWorldLightAmbient(0.25)
        sceneWidget.setBackgroundColor([0.1, 0.1, 0.3])
        sceneWidget.setCameraPosition(
            [-12.5, -7.0, 12.0],
            [1.57, 0.0, -1.04],
            [-0.01, -0.015, 1.7])

    def read_body_model(self, item, body_item, e1=""):
        """
        body モデルを読み込み配置する
        :param item:
        :param body_item:
        :param e1:
        :return:
        """
        folder = []
        try:
            # Number of body models.
            body_num = len(item[e1])
            for i in range(body_num):
                if not self.check_key("uri", item, e1, i):
                    folder_item = FolderItem()
                    folder_item.name = item[e1][i]["name"]
                    parent = item[e1][i]["parent"]
                    body_ret = self.check_parent_key(parent, body_item)
                    folder_ret = self.check_parent_key(parent, folder)
                    if parent == STR_WORLD_BODY_NAME:
                        # folder is NULL
                        self.world_item.addChildItem(folder_item)
                        folder.append(folder_item)
                    elif len(body_item) > 0 and body_ret[0]:
                        num = body_ret[1]
                        body_item[num].addChildItem(folder_item)
                        folder.append(folder_item)
                    elif len(folder) > 0 and folder_ret[0]:
                        # Otherwise
                        num = folder_ret[1]
                        folder[num].addChildItem(folder_item)
                        folder.append(folder_item)
                    else:
                        pass
                else:
                    # Generate BodyItem.
                    tmp = BodyItem()
                    # Load body model.
                    tmp.load(item[e1][i]["uri"])
                    tmp.name = item[e1][i]["name"]
                    # Display robot model.
                    tmp.setChecked(True)
                    body = tmp.body
                    if self.check_key("translation", item, e1, i):
                        # Check existence of "translation" key.
                        body.rootLink.setTranslation(item[e1][i]["translation"])
                    else:
                        # Otherwise
                        body.rootLink.setTranslation([0.0, 0.0, 0.0])

                    if self.check_key("rotation", item, e1, i):
                        # Check existence of "rotation" key.
                        r = math.radians(item[e1][i]["rotation"][0])
                        p = math.radians(item[e1][i]["rotation"][1])
                        y = math.radians(item[e1][i]["rotation"][2])
                    else:
                        # Otherwise
                        r = p = y = 0.0

                    body.rootLink.setRotation(rotFromRpy([r, p, y]))
                    body.calcForwardKinematics()
                    tmp.storeInitialState()

                    if not self.check_key("parent", item, e1, i) or self.item_list[e1][i]["parent"] == STR_WORLD_BODY_NAME:
                        # Check existence of "parent" key. Or "parent" key is "World"
                        self.world_item.addChildItem(tmp)
                    else:
                        # Otherwise
                        # Parent item name.
                        parent = self.item_list[e1][i]["parent"]
                        # Parent key existence check.
                        body_ret = self.check_parent_key(parent, body_item)
                        folder_ret = self.check_parent_key(parent, folder)
                        if len(body_item) > 0 and body_ret[0]:
                            # if it exists in the body item list.
                            num = body_ret[1]
                            body_item[num].addChildItem(tmp)
                        elif len(folder) > 0 and folder_ret[0]:
                            # if it exists in the folder item list.
                            num = folder_ret[1]
                            folder[num].addChildItem(tmp)
                        else:
                            # Otherwise
                            folder_item = FolderItem()
                            folder_item.name = self.item_list[e1][i]["parent"]
                            self.world_item.addChildItem(folder_item)
                            folder_item.addChildItem(tmp)
                            folder.append(folder_item)
                    # Add body item list.
                    body_item.append(tmp)
        except Exception as e:
            logger.error(e)
            print("[ERROR]", e)
        return

    @staticmethod
    def check_key(key, item, e1="", e2=-1):
        """
        item要素にkeyがあるか判定する
        :param key:
        :param item:
        :param e1:
        :param e2:
        :return:
        """
        try:
            if not e1:
                if not key in item:
                    return False
            elif e1 and e2 == -1:
                if not key in item[e1]:
                    return False
            else:
                if not key in item[e1][e2]:
                    return False
            return True
        except Exception as e:
            logger.error(e)
            print("[ERROR]", e)

    @staticmethod
    def check_parent_key(parent, body_item):
        """
        parent key があるか判定する
        :param parent:
        :param body_item:
        :return:
        """
        for i, item in enumerate(body_item):
            # 事前に name 属性があるか確認
            if hasattr(item, "name") and parent in item.name:
                return True, i
        return False, -1

    @staticmethod
    def get_world_item():
        """
        World要素を提供する
        :return:
        """
        world_item = None
        item = RootItem.instance.getChildItem()
        while item is not None:
            if item.name.lower() == STR_WORLD_BODY_NAME.lower():
                world_item = item
                break
            item = item.nextItem
        return world_item

    @staticmethod
    def get_child_item(parent_item, child_name):
        """
        要求された名称の要素を提供する
        :param parent_item:
        :param child_name:
        :return:
        """

        child_item = None
        item = parent_item.getChildItem()
        while item is not None:
            if item.name.lower() == child_name.lower():
                child_item = item
                break
            item = item.nextItem
        return child_item

    def get_free_space_item(self):
        """
        Free space要素を提供する
        :return:
        """
        free_space_item = None
        world_item = self.get_world_item()
        if world_item is None:
            logger.error("world_item is None")
            return free_space_item
        obstacle_item = self.get_child_item(world_item, STR_OBSTACLE)
        if obstacle_item is None:
            logger.error("obstacle_item is None")
            return free_space_item
        free_space_item = self.get_child_item(obstacle_item, STR_FREE_SPACE)
        return free_space_item

    def get_free_space_child_list(self):
        """
        Free space要素にある子要素をList型で提供する
        :return:
        """
        free_space_item_a = []
        free_space_item = self.get_free_space_item()
        free_space_child_item = free_space_item.getChildItem()
        while free_space_child_item is not None:
            free_space_item_a.append(free_space_child_item)
            free_space_child_item = free_space_child_item.nextItem
        return free_space_item_a

    def world_item_generator(self):
        """
        World要素を生成する
        """
        # Generate world item.
        RootItem.instance.addChildItem(self.world_item)

    def read_environment_model(self):
        """
        environmentモデルを読み込む
        :return:
        """
        try:
            self.read_body_model(self.item_list, self.env_item, STR_ENVIRONMENT)
        except Exception as e:
            logger.error(e)
            print("[Error]", e)

    def reallocate_free_space(self, update_data, body_name_a):
        """
        free spaceにモデルを再配置する
        :param update_data:
        :param body_name_a:
        :return:
        """
        self.remove_free_space_child(body_name_a)
        self.position_child_free_space(update_data, body_name_a, STR_ENVIRONMENT)

    def remove_free_space_child(self, delete_name_a):
        """
        free spaceの子要素削除
        :param delete_name_a:
        :return:
        """
        free_space_item_a = self.get_free_space_child_list()
        for del_item in free_space_item_a:
            detect = [view_object for view_object in delete_name_a if view_object.lower() in del_item.name.lower()]
            if len(detect) != 0:
                del_item.removeFromParentItem()
            if del_item.name.lower() == STR_END_BODY_NAME:
                del_item.removeFromParentItem()

    def position_child_free_space(self, item, view_a=None, e1=""):
        """
        渡されたitemをFree space要素に配置する
        :param item: item リスト
        :param view_a: 配置するBodyの名称
        :param e1: 要素名
        :return: 
        """
        try:
            free_space_item = self.get_free_space_item()
            if free_space_item is None:
                logger.error("free_space_item is None")
                return
            if view_a is None:
                view_a = []
            # Number of body models.
            for body_yaml in item[e1]:
                if not self.body_name_check(view_a, body_yaml):
                    continue
                if body_yaml.get("uri") is None:
                    continue
                # Generate BodyItem.
                body_child = self.create_body(body_yaml)
                if body_child is None:
                    continue
                free_space_item.addChildItem(body_child)
                body_child.setChecked(True)
            else:
                end_body = self.create_end_body()
                free_space_item.addChildItem(end_body)
                end_body.setChecked(True)

        except Exception as e:
            logger.error(e)
            print("[ERROR]", e)
        return

    @staticmethod
    def body_name_check(body_name_a, body_yaml):
        """

        :param body_name_a:
        :param body_yaml:
        :return:
        """
        result = False
        applicable_body_a = [body_name for body_name in body_name_a if body_name.lower() in body_yaml["name"].lower()]
        if len(applicable_body_a) > 0:
            result = True
        return result

    @staticmethod
    def create_end_body():
        """

        :return:
        """
        body_item = BodyItem()
        # Load body model.
        body_item.load("${SHARE}/LICTiA/model/dummy.body")
        body_item.setName(STR_END_BODY_NAME)
        body_item.body.rootLink.setTranslation([0.0, 0.0, -1])
        body_item.body.rootLink.setRotation(rotFromRpy([0.0, 0.0, 0.0]))
        body_item.body.calcForwardKinematics()
        body_item.storeInitialState()
        return body_item

    @staticmethod
    def create_body(set_data):
        """

        :param set_data:
        :return:
        """
        body_item = BodyItem()
        # Load body model.
        body_item.load(set_data["uri"])
        body_item.setName(set_data["name"])

        translation = [0.0, 0.0, 0.0]
        if set_data.get("translation") is not None:
            translation = set_data["translation"]
        body_item.body.rootLink.setTranslation(translation)

        r = p = y = 0.0
        if set_data.get("rotation") is not None:
            r = math.radians(set_data["rotation"][0])
            p = math.radians(set_data["rotation"][1])
            y = math.radians(set_data["rotation"][2])

        body_item.body.rootLink.setRotation(rotFromRpy([r, p, y]))
        body_item.body.calcForwardKinematics()
        body_item.storeInitialState()
        return body_item
